(defmeth mmr-model-object-proto :visualize ()
  (if (not (eq current-object self)) (setcm self))
  (when (= 1 (send self :num-iv))
        (error "Visualization not possible when there is only one predictor. Use univariate Regression Analysis."))
  (let* ((cpoints (center (select (send self :data-matrix)
                                  (iseq (send self :nobs))
                                  (send self :iv))))
         (rays (transpose (send self :beta)));(rays (send self :coefs))
         (nredun (send self :redundancy))
         (redun-rays (if nredun
          (matmult 
           (diagonal  (select (send self :redun-evals) (iseq nredun)))
           (transpose (select (send self :redun-coefs)
                              (iseq (send self :num-iv)) (iseq nredun))))))
         (numerator (max (mapcar #'ssq (row-list cpoints))))
         (denom (max (mapcar #'ssq (row-list rays))))
         (scale-factor (sqrt (/ numerator denom)))
         
         (spin-dim (min 9 (send self :num-iv)))
         (scaled-coefs (* scale-factor
                          (select (column-list rays) (iseq spin-dim))))
         (redlabels nil)
         (denom-redun nil)
         (scale-factor-redun nil)
         (scaled-redun-coefs nil)
         (ray-labels (select (send self :variables) (send self :dv))))
    (when redun-rays
          (setf redlabels
                (mapcar #'(lambda (x) (format nil "Red~d" x))
                        (iseq nredun)))
          (setf denom-redun 
                (max (mapcar #'ssq (append (row-list rays)
                                           (row-list redun-rays)))))
          (setf scale-factor-redun (sqrt (/ numerator denom-redun)))
          (setf scaled-redun-coefs 
                (* scale-factor-redun
                   (select (column-list redun-rays) (iseq spin-dim)))))
    (when (> (send self :num-iv) 2)
          (spreadplot (column-list cpoints)
                      (send self :labels) 
                      (select (send self :variables) (send self :iv))
                      :scale-type 'fixed
                      :show-histogram nil
                      :show-reg-plot t)
          (send spin-plot   :depth-cuing nil)
          #+color(send spin-plot   :point-color 
                       (iseq (send spin-plot :num-points)) 'green)
          (send spin-plot   :point-symbol
                (iseq (send spin-plot :num-points)) 'square)
          (send spin-plot   :add-rays scaled-coefs 
                :ray-labels ray-labels :ray-color 'blue)
          (when redun-rays 
                (send spin-plot :add-rays scaled-redun-coefs 
                      :ray-labels redlabels :ray-color 'red))
          (send list-obs :title "Labels")
          (send spin-plot :title "Spin-BiPlot")
          (send (send spin-plot :menu) :title "Spin-BiPlot")
          (send scatterplot :title "BiPlot")
          (send (send scatterplot :menu) :title "BiPlot")
          (send scatmat :title "Predictor Plots")
          (send spin-plot :redraw)
          (setf reg-plot 
                (plot-points (col (send self :scores) 0)
                             (send (select (send self :reg-models) 0) :y)
                             :show nil :location location22 :size graph-size))
          
          (send reg-plot :title "Regression Plot")
          (send reg-plot :new-menu "RegPlot" 
              :items '(LINK SHOWING-LABELS MOUSE RESIZE-BRUSH DASH 
                            ERASE-SELECTION FOCUS-ON-SELECTION SHOW-ALL 
                            SYMBOL COLOR SELECTION))
          (defmeth reg-plot :close ()
            (send scatmat :close))
          (defmeth reg-plot :reg-plot (i mob) 
            (let* ((x (send (select (send mob :reg-models) i) :y))
                   (y (col (send mob :scores) i))
                   (n (length y))
                   (symbols (send self :point-symbol (iseq n)))
                   (colors  (send self :point-color (iseq n)))
                   (states  (send self :point-state (iseq n)))
                   (min (min (combine x y)))
                   (max (max (combine x y)))
                   (gnr (get-nice-range min max 5))
                   (axmin (first gnr))
                   (axmax (second gnr))
                   )
              (send reg-plot :range 0 (first gnr) (second gnr) :draw nil)
              (send reg-plot :range 1 (first gnr) (second gnr) :draw nil)
              (send reg-plot :x-axis t t (third gnr) :draw nil)
              (send reg-plot :y-axis t t (third gnr) :draw nil)

              (send reg-plot :clear)
              (send reg-plot :add-points x y 
                    :point-labels (send mob :labels) :draw nil)
              (send reg-plot :point-symbol (iseq n) symbols :draw nil)
              (send reg-plot :point-color  (iseq n) colors :draw nil)
              (send reg-plot :point-state  (iseq n) states :draw nil)
              (send reg-plot :variable-label 0
                    (select (select (send mob :variables) 
                                    (send mob :dv)) i))
              (send reg-plot :variable-label 1 
                    (strcat "Predicted " 
                            (select (select (send mob :variables) 
                                            (send mob :dv)) i)))
            
              (send reg-plot :add-lines 
                    (list (list axmin axmax) (list axmin axmax)) :draw nil)
            
              (send reg-plot :redraw)))

          (defmeth list-obs :do-select-click (x y m1 m2)
            (call-next-method x y m1 m2)
            (when (send self :selection)
                  (let* ((nsel (first (send self :selection)))
                         (nobs (send *current-model* :nobs))
                         (ndv  (length (send *current-model* :dv)))
                         )
                    (when (and (not (< nsel nobs))
                               (< nsel (+ nobs ndv)))
                          (send reg-plot :reg-plot (- nsel nobs) 
                                current-model)
                          )
                    )
                  ))
          
          (send reg-plot :reg-plot 0 self)
          (send reg-plot :show-window)
          (send reg-plot :showing-labels t)
          (send reg-plot :linked t)
          #+color(send reg-plot :use-color t)
          #+color(send reg-plot :point-color 
                       (iseq (send reg-plot :num-points)) 'green)
          (send reg-plot :point-symbol
                (iseq (send reg-plot :num-points)) 'square)
          )
    (when (= (send self :num-iv) 2)
          (setf scatterplot (plot-points 
                    (column-list cpoints)
                    :title (send self :title)
                    :location location11
                    :size (+ (/ menu-bar-height 2) (* 2 graph-size))
                    :point-labels (send self :labels)
                    :variable-labels 
                    (select (send self :variables) (send self :iv))
                    :show nil))
          (send scatterplot :showing-labels t)
          (send scatterplot :linked t)
          (send scatterplot :new-menu "ScatPlot" 
              :items '(LINK SHOWING-LABELS MOUSE RESIZE-BRUSH DASH 
                            ERASE-SELECTION FOCUS-ON-SELECTION SHOW-ALL 
                            SYMBOL COLOR SELECTION))
          (send scatterplot :redraw)
          (send scatterplot :show-window)
          (defmeth scatterplot :close ()
            (send list-obs :remove)
            (call-next-method))
          (setf list-obs (send current-object :obs-namelist))
          (send list-obs  :linked t)
          (defmeth list-obs :close ()
            (send scatterplot :close))
          (send list-obs :new-menu "Labels" 
                :items '(LINK MOUSE DASH ERASE-SELECTION FOCUS-ON-SELECTION 
                              SHOW-ALL COLOR SELECTION))
          )
    #+color(send scatterplot :use-color t)
    #+color(send scatterplot :point-color 
                 (iseq (send scatterplot :num-points)) 'green)
    (send scatterplot :point-symbol
          (iseq (send scatterplot :num-points)) 'square)
    (send scatterplot :add-rays scaled-coefs 
          :ray-labels ray-labels :ray-color 'blue)
    (send list-obs :add-points (send self :num-dv) 
          :point-labels ray-labels)
    (when redun-rays 
          (send scatterplot :add-rays scaled-redun-coefs 
                :ray-labels redlabels :ray-color 'red)
          (send list-obs :add-points (send self :redundancy)
                      :point-labels redlabels))
    (send scatterplot :redraw)
    t))

(provide "mmrvis")